/*************************************************************************
 * clewd.c Copy load eject wait disc utility, v 0.2
 *
 * Copyright (c) 2004 George Shuklin, http://shounen.ru/soft/clewd/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *   
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *   
 * You should have received a copy of the GNU General Public License along
 * with this program (see the file COPYING); if not, write to the Free 
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 * MA  02111-1307  USA  or visit http://www.gnu.org/copyleft/gpl.html
 *
 *************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <winioctl.h>
#include <conio.h>
#include <direct.h>
#include <errno.h>

#define DISC_PING_DELAY 5000
/*delay between drive rechecking (when waiting), in ms*/
#define KEYBOARD_POLL 333
/*keyboard checking (for user waiting canceling), in ms*/

enum{
    ACTION_INVERT=0,
	ACTION_EJECT=1,
	ACTION_LOAD=2,
	ACTION_WAIT_FOR_LOAD=3,
	ACTION_MKDIR_FROM_LABEL_AND_COPY=4,
	ACTION_CHECK=5,
	ACTION_COPY=6
};/* values of config.action*/

typedef struct{
	int action;
	char drive_letter; /*0 means unknown (error), 1 request for autodetect first cd/dvd (2 - for second  3 for fird, supports up to 26 drives)*/
	char* drive_path; /*path to drive in form \\.\d: , setup within SelectDrive, allocating in Config */
	int silence; /*if !0 - do not print any, if >2 do not make a noise on drive errors*/
	char* label; /*disc label - fill in GetLabel*/
}config_t;


void ExitErr(config_t* config, const char* message, const int code){
	if(!config->silence){
   		fprintf(stderr,"error %d: %s\n",code,message);
		fprintf(stderr,message);
	}
	exit(code);
}

void Help_n_Exit(){
	printf(
			"Clewd - Copy Load/Eject Wait for Disc application v0.1\n"
			"usage: clewd [-command] [drive:|number]\n"
			"commands:\n"
			"\t-e - eject\n"
			"\t-l - load\n"
			"\t-i - load, if no disc, eject if disc in drive (default bihavior)\n"
			"\t-w - wait until disc inserted (or user press any key)\n"
			"\t-t - test disc status (no disc/disc avaible)\n"
			"\t-cl - create a dir with name of disc label and execute xcopy command (see README)\n"
			"\t-c - copy files from disc to current dir\n"
			"\t-s - do not print anything on screen (ony do and set errorlevels)\n"
			"\t-h - mmm... do something\n"
			"drive - a:, b:, c:, ..., z: - specify a driveletter\n"
			"number - 1, 2, ..., 26 - autodetect a first/second/etc CD/DVD drive\n"
			"by default (if no other specified) used first avaible optical drive\n"
			"\tStatus: 0: -l, -e, -i -c - operation sucsessfull\n"
			"\t\t   -t -w - disc inserted\n"
			"\t\tnot zero: -e -l -i -c -cl, -tl - opearion fail\n"
			"\t\t   -t - no CD\n"
			"\t\t   -w - user chancel waiting by pressing a key\n"
			"(c) George Shuklin,2004, gs@shounen.ru http://shounen.ru/soft/clewd/"
		);
	exit(0);
}


int Configure(int argc, char* argv[], config_t* config){
	int c;
	config->silence=0;
	config->action=ACTION_INVERT;
	config->drive_letter=1;
	config->drive_path=strdup("\\\\.\\$:"); 
	config->label=NULL;
	for(c=1;c<argc;c++){
		if(argv[c][0]=='-'){
			switch(argv[c][1]){
				case 'e': config->action=ACTION_EJECT; break;
				case 'l': config->action=ACTION_LOAD; break;
				case 'c': 
					if(argv[c][2]=='l') config->action=ACTION_COPY;
					if(!argv[c][2])config->action=ACTION_MKDIR_FROM_LABEL_AND_COPY; 
					break;
				case 'w': config->action=ACTION_WAIT_FOR_LOAD; break;
				case 'i': config->action=ACTION_INVERT;break;
				case 't': config->action=ACTION_CHECK;break;
				case 's': config->silence=1; break;
				case 'h': Help_n_Exit();
				case '-':
						/*TODO*/
						Help_n_Exit();
				default:
					fprintf(stderr,"unknown option %d: %s\n",c,argv[c]);
					return 10;
			}
		}
		else{
			if(argv[c][1]!=':'&&argv[c][1]){
				fprintf(stderr,"unknown option %d: %s\n",c,argv[c]);
				return 10;
			}
			if(argv[c][0]<='9'&&argv[c][0]>='1') /*zero not valid*/ 
				config->drive_letter=argv[c][0]-'0';
			else 
				config->drive_letter=argv[c][0];

		}
    }
	return 0;
};


char AutoDetectDrive (config_t* config, int order_number){
       char testline[]="\\\\.\\$:\\";
       int u;
       int c;
       int counter=0;
       for(c='A';c<='Z';c++){
       		testline[4]=c;
			u = GetDriveType(testline);
			if(u==DRIVE_CDROM) counter++; /*DVD-ROMs has some type as CD*/
			if(counter==order_number) {
				if(!config->silence) printf("drive %c: ",c);
				return c;
			}
       }
	return 0;
}

void SelectDrive(config_t *config){
	if(!config->drive_letter) ExitErr(config,"drive letter unknown", 11);
	if(config->drive_letter<27){
		config->drive_letter=AutoDetectDrive(config, config->drive_letter);
		if(!config->drive_letter) ExitErr(config,"autodetect failed",13);
	}
	config->drive_path[4]=config->drive_letter;
};

inline int WorkMessage(config_t* config,const char* message,const int result){
	if(!config->silence){
		printf("%s: %s\n",message,result?"OK":"FAIL");
	}
	return !result;
}


int DoAction(config_t* config, int type){
	HANDLE h = CreateFile(   config->drive_path, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING,0, NULL );
	int res=0;
    DWORD dwBytesReturned;
	if(h==INVALID_HANDLE_VALUE) return 0;
	res=DeviceIoControl( h,type,NULL, 0,NULL, 0,&dwBytesReturned,NULL);
	CloseHandle(h);
	return res;
}


inline int EjectDrive(config_t* config){
	return DoAction(config, IOCTL_DISK_EJECT_MEDIA);
}

inline int LoadDrive(config_t* config){
	return DoAction(config, IOCTL_DISK_LOAD_MEDIA);
}

inline int DriveStatus(config_t* config){
	return DoAction(config, IOCTL_DISK_CHECK_VERIFY);
}

int DriveWait(config_t* config){
	int poll_counter;
    if(DriveStatus(config)) return 0; /*ok*/
	if(!config->silence) printf("Please insert disc into  %c:  drive or press any key to exit\n", config->drive_letter);
	do{
		poll_counter=0;
	    do{
	    	poll_counter+=KEYBOARD_POLL;
	    	sleep(KEYBOARD_POLL);
   			if(_kbhit()) {
				getch();/*just remove pressed key from stdin*/
				return 15; /*error - user break*/
			}
		}while (poll_counter<DISC_PING_DELAY);
	}while(!DriveStatus(config));
	return 0;
}


int GetDriveLabel(config_t* config){
    #define LABEL_BUFFER_SIZE 511
	char buffer[LABEL_BUFFER_SIZE+1];
	char name[]="\\\\.\\$:\\";
	int result;
	name[4]=config->drive_letter;
	result=GetVolumeInformation(name,  buffer,LABEL_BUFFER_SIZE, NULL, NULL, NULL, NULL, 0 );
   buffer[LABEL_BUFFER_SIZE]=0;
   if(result) config->label=strdup(buffer);
   return result;
}


void MakeDirFromLabel(config_t* config){
	int status;
	if(!mkdir(config->label))
		if(chdir(config->label)) ExitErr(config,"error create/change dir",14);

}

int CopyAllFiles(config_t* config){
	char execute_line[512]="xcopy /s /y $:\\*.* .";
	if(config->drive_letter<32) return 0;
	if(execute_line[12]=='$') execute_line[12]=config->drive_letter;
	else ExitErr(config,"internal error within CopyAllLine function",16);
	if(config->silence){
		strcat(execute_line, "1> NUL 2>NUL");
	}else{
		printf("(Press Ctrl-C to terminate)\nexecute copying: %s\n",execute_line);
	}
	return system(execute_line);
}

int main(int argc, char* argv[]){
	config_t config;
	HANDLE h;
	Configure(argc, argv, &config);
	SelectDrive(&config);
	switch(config.action){
		case ACTION_EJECT: return WorkMessage(&config,"eject",EjectDrive(&config)); break;
		case ACTION_LOAD:  return WorkMessage(&config,"load",LoadDrive(&config)); break;
		case ACTION_INVERT: 
						if(DriveStatus(&config))
							return WorkMessage(&config,"eject", EjectDrive(&config));
						else 
							return WorkMessage(&config,"load", LoadDrive(&config));
						 break;
   		case ACTION_WAIT_FOR_LOAD: 
   						return WorkMessage(&config,"result ",!DriveWait(&config));/*todo*/
						break;
   		case ACTION_CHECK: return WorkMessage(&config, "satus ",DriveStatus(&config));
		case ACTION_MKDIR_FROM_LABEL_AND_COPY: 
						if(GetDriveLabel(&config)){
							if(!config.silence) printf("Label: %s\n",config.label);
							MakeDirFromLabel(&config);
						}else ExitErr(&config,"no label (or no CD)",12);
						/*passthrough*/
		case ACTION_COPY:
				if(DriveStatus(&config))
					return CopyAllFiles(&config);
        			else 
					ExitErr(&config, "nothing to copy (no CD)\n",17);
	}
	return 0;
}
